__version__ = "1.2.5.dev0"
__package_name__ = "skin-tone-classifier"
__app_name__ = "Skin Tone Classifier"
__description__ = "An easy-to-use library for skin tone classification"
__author__ = "Chenglong Ma"
__author_email__ = "chenglong.m@outlook.com"
__author_website__ = "https://chenglongma.com/"
__url__ = "https://chenglongma.com/SkinToneClassifier/"
__code__ = "https://github.com/ChenglongMa/SkinToneClassifier/"
__issues__ = "https://github.com/ChenglongMa/SkinToneClassifier/issues"
__license__ = "GPLv3"
__copyright__ = "2022"
